;;;
;;; Time-stamp: <2008-01-24 11:35:46 nhw>
;;;
;;; Copyright (C) by Noel Welsh. 
;;;

;;; This library is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU Lesser
;;; General Public License as published by the Free Software
;;; Foundation; either version 2.1 of the License, or (at
;;; your option) any later version.

;;; This library is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE.  See the GNU Lesser General Public
;;; License for more details.

;;; You should have received a copy of the GNU Lesser
;;; General Public License along with this library; if not,
;;; write to the Free Software Foundation, Inc., 59 Temple
;;; Place, Suite 330, Boston, MA 02111-1307 USA

;;; Author: Noel Welsh <noelwelsh@yahoo.com>
;;
;;
;; Commentary:

;; Basic operators for matrices
;;
;; Types:
;;   vector : synonym for f64vector
;;
;; Naming convention
;;   m prefix indicates operates on matrix
;;   / suffix indicates a fold
;;   . suffix indicates operates element-wise
;;   : suffix indicates operates column-wise
;;   - suffix indicates operates row-wise
(module matrix mzscheme

  (require
   (lib    "etc.ss")
   (lib    "for.ss")
   (planet "srfi-4-comprehensions.ss" ("wmfarr" "srfi-4-comprehensions.plt" 1))
   (planet "all.ss" ("wmfarr" "plt-linalg.plt" 1))
   (file   "vector.ss"))
  
  (provide matrix
           (all-defined))

  ;; m? : any -> boolean
  (define m? matrix?)
  
  ;; m= : matrix matrix [epsilon] -> (U #t #f)
  (define m=
    (opt-lambda (m1 m2 [epsilon 0.0])
      (and (= (matrix-rows m1)
              (matrix-rows m2))
           (= (matrix-cols m1)
              (matrix-cols m2))
           (every?-ec
            (:parallel (:matrix x1 m1)
                       (:matrix x2 m2))
            (<= (abs (- x1 x2)) epsilon)))))

  (define (display-matrix m)
    (printf "#matrix(~a ~a~n" (mrows m) (mcols m))
    (do-ec (:range r (mrows m))
           (begin
             (display "        ")
             (do-ec (:range c (mcols m))
                    (:let x (mref m r c))
                    (printf "~a " x))
             (newline)))
    (display ")"))
  
  ;; m : rows cols number ... -> matrix
  (define (m rows cols . vals)
    (matrix-ec rows cols (:list v vals) v))

  (define mrows matrix-rows)
  (define mcols matrix-cols)
  (define mref matrix-ref)
  (define mset! matrix-set!)
  (define mtranspose matrix-transpose)
  
  ;; mmake : rows cols fill
  ;;
  ;; Constructor
  (define mmake make-matrix)

  ;; v->m : integer integer vector [boolean] -> matrix
  ;;
  ;; Creating a matrix by repeating vector rows times in the
  ;; rows dimension and cols time in the column dimension
  ;;
  ;; Assumes the vector is a column vector unless the final
  ;; optional argument is false
  (define v->m
    (opt-lambda (rows cols vector [col? #t])
      (let ([length (vlength vector)])
        (if col?
            (matrix-ec (* length rows) cols
                       (:range c cols)
                       (:range r rows)
                       (:f64vector elt vector)
                       elt)
            (matrix-ec rows (* length cols) 
                       (:range c cols)
                       (:f64vector elt vector)
                       (:range r rows)
                       elt)))))

  ;; mzeros : integer integer -> matrix
  (define (mzeros rows cols)
    (make-matrix rows cols 0.0))
  
  ;; m- : matrix matrix -> matrix
  (define m- matrix-sub)

  ;; m+ : matrix matrix -> matrix
  (define m+ matrix-add)

  ;; m* : matrix matrix -> matrix
  (define m* matrix-mul)

  ;; m*. : matrix matrix -> matrix
  ;;
  ;; Element-wise multiplication
  (define (m*. m1 m2)
    (matrix-ec
     (mrows m1) (mcols m1)
     (:parallel (:matrix x1 m1)
                (:matrix x2 m2))
     (* x1 x2)))
     
  ;; m/s : matrix real -> matrix
  (define (m/s m s)
    (matrix-ec (mrows m) (mcols m)
               (:matrix x m)
               (/ x s)))

  ;; m+s : matrix real -> matrix
  (define (m+s m s)
    (mmap (lambda (x) (+ x s)) m))
  
  ;; m*v : matrix vector -> matrix
  (define m*v matrix-f64vector-mul)
  
  ;; mmap : (real -> real) matrix -> matrix
  (define (mmap fn m)
    (matrix-ec (matrix-rows m) (matrix-cols m)
               (:matrix x m)
               (fn x)))

  ;; m*s : matrix real -> matrix
  (define (m*s m s)
    (mmap (lambda (x) (* x s)) m))
  
  ;; m/:v : matrix vector -> matrix
  (define (m/:v m v)
    (let ([rows (matrix-rows m)]
          [cols (matrix-cols m)])
      (matrix-ec rows cols
                 (:range c cols)
                 (:f64vector v1 (index r) v)
                 (:let m1 (mref m r c))
                 (/ m1 v1))))

  (define (m/-v m v)
    (let ([cols (matrix-cols m)]
          [rows (matrix-rows m)])
      (matrix-ec rows cols
                 (:matrix m1 (index r c) m)
                 (:let v1 (vref v c))
                 (/ m1 v1))))
  
  ;; m*-v : matrix vector -> matrix
  ;;
  ;; Multiply the rows of m by v
  (define (m*-v m v)
    (let ([rows (matrix-rows m)]
          [cols (matrix-cols m)])
      (matrix-ec rows cols
                 (:matrix x (index r c) m)
                 (* x (vref v c)))))

  ;; m*:v : matrix vector -> matrix
  (define (m*:v m v)
    (let ([rows (matrix-rows m)]
          [cols (matrix-cols m)])
      (matrix-ec rows cols
                 (:matrix x (index r c) m)
                 (* x (vref v r)))))

  ;; m+/ : matrix -> vector
  ;;
  ;; Return the sum of all the elements in the given matrix
  (define (m+/ m)
    (sum-ec
     (:matrix x m)
     x))
  
  ;; m+/- : matrix -> vector
  ;;
  ;; Returns a vector containing the sum of the rows of the given matrix
  (define (m+/- m)
    (let* ([cols (matrix-cols m)])
      (matrix-f64vector-mul m (vones cols)))) 

  ;; m+/: : matrix -> vector
  ;;
  ;; Returns a vector containing the sum of the columns of the given matrix
  (define (m+/: m)
    (let* ([rows (matrix-rows m)])
      (f64vector-matrix-mul (vones rows) m)))
  
  ;; mcolumn : matrix integer -> vector
  ;;
  ;; Return a vector representing the values in the matrix
  ;; of the given column
  (define (mcolumn m col)
    (let ([rows (matrix-rows m)])
      (f64vector-of-length-ec rows
       (:range r rows)
       (matrix-ref m r col))))

  ;; mrow : matrix integer -> vector
  (define (mrow m row)
    (let ([cols (mcols m)])
      (f64vector-of-length-ec cols
       (:range c cols)
       (matrix-ref m row c))))
  
  ;; mset-column! : matrix integer vector -> void 
  (define (mset-column! matrix col vec)
    (do-ec
     (:range r (vlength vec))
     (matrix-set! matrix r col (vref vec r))))

  ;; mset-row! : matrix integer vector -> void 
  (define (mset-row! matrix row vec)
    (do-ec
     (:f64vector i (index c) vec)
     (matrix-set! matrix row c i)))

  ;; syntax for/fold/mcols : (natural natural) (var ...) (gen ...) expr)
  (define-syntax (for/fold/mcols stx)
    (syntax-case stx ()
      [(for/fold/mcols (r c) ([accum-id init-expr] ...) ([id seq-expr] ...) body ...)
       (with-syntax ([(shadow-id ...)
                      (datum->syntax-object
                       stx
                       (map gensym (syntax-object->datum (syntax (accum-id ...)))))])
         (syntax
          (let ([m (mmake r c 0.)]
                [accum-id init-expr] ...)
            (for ([column (in-range c)] [id seq-expr] ...)
                 (let-values ([(col shadow-id ...)
                               (begin body ...)])
                   (set! accum-id shadow-id) ...
                   (mset-column! m column col)))
            m)))]))

  ;; syntax for/fold/mcolsi : (natural natural) (var ...) (gen ...) expr)
  (define-syntax (for/fold/mcolsi stx)
    (syntax-case stx ()
      [(for/fold/mcols (r c) (col-id col-expr) ([accum-id init-expr] ...) ([id seq-expr] ...) body ...)
       (with-syntax ([(shadow-id ...)
                      (datum->syntax-object
                       stx
                       (map gensym (syntax-object->datum (syntax (accum-id ...)))))])
         (syntax
          (let ([m (mmake r c 0.)]
                [accum-id init-expr] ...)
            (for ([col-id col-expr] [id seq-expr] ...)
                 (let-values ([(col shadow-id ...)
                               (begin body ...)])
                   (set! accum-id shadow-id) ...
                   (mset-column! m col-id col)))
            m)))]))

  ;; mfilter : (real -> real) matrix -> matrix
  (define (mfilter f m)
    (matrix-ec
     (mrows m) (mcols m)
     (:matrix x m)
     (f x)))

  )
