;;;
;;; Time-stamp: <2008-01-14 11:32:44 noel>
;;;
;;; Copyright (C) by Noel Welsh. 
;;;

;;; This library is free software; you can redistribute it
;;; and/or modify it under the terms of the GNU Lesser
;;; General Public License as published by the Free Software
;;; Foundation; either version 2.1 of the License, or (at
;;; your option) any later version.

;;; This library is distributed in the hope that it will be
;;; useful, but WITHOUT ANY WARRANTY; without even the
;;; implied warranty of MERCHANTABILITY or FITNESS FOR A
;;; PARTICULAR PURPOSE.  See the GNU Lesser General Public
;;; License for more details.

;;; You should have received a copy of the GNU Lesser
;;; General Public License along with this library; if not,
;;; write to the Free Software Foundation, Inc., 59 Temple
;;; Place, Suite 330, Boston, MA 02111-1307 USA

;;; Author: Noel Welsh <noelwelsh@yahoo.com>
;;
;;
;; Commentary:

;; Basic operators for f64vectors
;;
;; Types:
;;   vector : synonym for f64vector
;;
;; Naming convention
;;   v prefix indicates operates on vector
;;   / suffix indicates a fold
;;   . suffix indicates operates element-wise
(module vector mzscheme

  (require
   (lib "etc.ss")
   (planet "srfi-4-comprehensions.ss" ("wmfarr" "srfi-4-comprehensions.plt" 1))
   (planet "all.ss" ("wmfarr" "plt-linalg.plt" 1)))

  (provide
   (all-defined))

  ;; v= : vector vector [real] -> (U #t #f)
  (define v=
    (opt-lambda (v1 v2 [epsilon 0.0])
      (and (eq? (f64vector-length v1)
                (f64vector-length v2))
           (every?-ec
            (:parallel (:f64vector x1 v1)
                       (:f64vector x2 v2))
            (<= (abs (- x1 x2)) epsilon)))))

  ;; v? : any -> boolean
  (define v? f64vector?)
  
  (define (display-vector vector)
    (display "#f64vector(")
    (do-ec (:f64vector x vector)
           (printf "~a " x))
    (display ")"))

  (define (vector->string vector)
    (let ([op (open-output-string)])
      (parameterize ([current-output-port op])
        (display-vector vector))
      (get-output-string op)))
  
  ;; v : number ... -> vector
  ;;
  ;; Constructor
  (define (v . elts)
    (apply f64vector (map exact->inexact elts)))

  ;; vmake : length fill -> vector
  ;;
  ;; Constructor
  (define vmake make-f64vector)

  ;; vzeros : integer -> vector
  (define (vzeros length) (vmake length 0.0))

  ;; vcopy : vector -> vector
  (define vcopy f64vector-copy)
  
  ;; vlength : vector -> integer
  (define vlength f64vector-length)

  ;; vref : vector integer -> real
  (define vref f64vector-ref)

  ;; vset! : vector integer real -> void
  (define vset! f64vector-set!)
  
  ;; v+/ : vector -> real
  (define (v+/ v)
    (sum-ec (:f64vector x v) x))

  ;; v* : vector vector -> real
  ;;
  ;; Euclidean dot product of two vector
  (define v* f64vector-dot)

  ;; v- : vector vector -> vector
  (define v- f64vector-sub)

  ;; v+ : vector vector -> vector
  (define v+ f64vector-add)
  
  ;; v*s : vector real -> vector
  ;;
  ;; Vector scale by a scalar
  (define v*s f64vector-scale)

  ;; v*m : vector matrix -> vector
  (define v*m f64vector-matrix-mul)

  ;; v+s : vector real -> vector
  (define (v+s v s)
    (f64vector-of-length-ec
     (f64vector-length v)
     (:f64vector x v)
     (+ x s)))
  
  ;; v-s : vector real -> vector
  ;;
  ;; Subtract a scalar from every element of a vector
  (define (v-s v s)
    (f64vector-of-length-ec
     (f64vector-length v)
     (:f64vector x v)
     (- x s)))

  ;; v/s : vector real -> vector
  (define (v/s v s)
    (f64vector-of-length-ec
     (f64vector-length v)
     (:f64vector x v)
     (/ x s)))
  
  ;; vmap : (real -> real) vector -> vector
  (define (vmap fn v)
    (f64vector-of-length-ec
     (f64vector-length v)
     (:f64vector x v)
     (fn x)))

  ;; vmapi : (integer real -> real) vector -> vector
  (define (vmapi fn v)
    (f64vector-of-length-ec
     (f64vector-length v)
     (:f64vector x (index i) v)
     (fn i x)))

  ;; vfold : (real 'a -> 'a) 'a vector -> 'a
  (define (vfold fn seed v)
    (let ([accum seed])
      (do-ec
       (:f64vector x v)
       (set! accum (fn x accum)))
      accum))

  ;; vfoldi : (integer real 'a -> 'a) 'a vector -> 'a
  (define (vfoldi fn seed v)
    (let ([accum seed])
      (do-ec
       (:f64vector x (index i) v)
       (set! accum (fn i x accum)))
      accum))

  
  ;; v*. : vector vector -> vector
  (define (v*. v1 v2)
    (f64vector-of-length-ec
     (f64vector-length v1)
     (:parallel (:f64vector x1 v1)
                (:f64vector x2 v2))
     (* x1 x2)))

  ;; vones : integer -> vector
  ;;
  ;; Create a vector with elements equal to one
  (define (vones length)
    (make-f64vector length 1.0))

  ;; vnormalise : f64vector -> f64vector
  ;;
  ;; Normalise values of a vector so they sum to one
  (define (vnormalise v)
    (v/s v (v+/ v)))

  ;; vmax : f64vector -> real
  ;;
  ;; Maximum element in the vector.  Returns -inf.0 for the
  ;; empty vector
  (define (vmax v)
    (vfold (lambda (x max)
             (if (> x max) x max))
           -inf.0
           v))

  ;; vmaxi : f64vector -> natural
  ;;
  ;; Index of the maximum element in the vectors.  Returns
  ;; -1 for the empty vector
  (define (vmaxi v)
    (cdr 
     (vfoldi (lambda (i x max)
               (let ([max-v (car max)]
                     [max-i (cdr max)])
                 (if (> x max-v) (cons x i) max)))
             (cons -inf.0 -1)
             v)))

)
     
  
